using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using UnityEngine;

namespace CheesyFX
{
    public static class KeyInput
    {
        public static JSONStorableFloat doubleTapTimeout = new JSONStorableFloat("Double Tap Timeout", .3f, .1f, 1f);
        public static Action<KeyCode, KeyCode> onSingle;
        public static Action<KeyCode> onDouble;
        public static Action<KeyCode> onDuringSingleStart;
        public static Action<KeyCode> duringSingleDown;
        public static Action<KeyCode> onDuringSingleEnd;
        public static Action<KeyCode> onDuringDoubleStart;
        public static Action<KeyCode> duringDoubleDown;
        public static Action<KeyCode> onDuringDoubleEnd;

        private static List<KeyCode> modifiers = new List<KeyCode>
        {
            KeyCode.LeftControl,
            KeyCode.RightControl,
            KeyCode.LeftShift,
            KeyCode.RightShift,
            KeyCode.LeftAlt,
            KeyCode.RightAlt
        };

        private static Dictionary<KeyCode, int> tapCounts = new Dictionary<KeyCode, int>();

        public static void OnKeyDown(KeyCode key)
        {
            if (!tapCounts.Keys.Contains(key))
            {
                tapCounts[key] = 1;
                TapInit(key).Start();
            }
            else tapCounts[key] += 1;
        }

        static IEnumerator TapInit(KeyCode key)
        {
            KeyCode modifier = modifiers.FirstOrDefault(Input.GetKey);
            yield return new WaitForSeconds(doubleTapTimeout.val);
            int tapCount = tapCounts[key];
            if (Input.GetKey(key))
            {
                if (tapCount == 1)
                {
                    onDuringSingleStart?.Invoke(key);
                    if(duringSingleDown != null) DuringSingleDown(key).Start();
                }

                if (tapCount > 1)
                {
                    onDuringDoubleStart?.Invoke(key);
                    if(duringDoubleDown != null) DuringDoubleDown(key).Start();
                }
            }
            else
            {
                if (tapCount == 1 && onSingle != null) onSingle(key, modifier);
                if (tapCount > 1 && onDouble != null) onDouble(key);
                tapCounts.Remove(key);
            }
        }

        private static IEnumerator DuringSingleDown(KeyCode key)
        {
            while (Input.GetKey(key))
            {
                // "During Single Down".Print();
                duringSingleDown(key);
                yield return new WaitForFixedUpdate();
            }
            onDuringSingleEnd?.Invoke(key);
            tapCounts.Remove(key);
        }
        
        private static IEnumerator DuringDoubleDown(KeyCode key)
        {
            while (Input.GetKey(key))
            {
                // "During Double Down".Print();
                duringDoubleDown(key);
                yield return new WaitForFixedUpdate();
            }
            onDuringDoubleEnd?.Invoke(key);
            tapCounts.Remove(key);
        }
    }
}